/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: BenchWorld.cs
// Description: Abstracts the BenchWorld - the underlying data structure that
//              represents the matrix so that it can be changed.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 22.01.2008
//
//-----------------------------------------------------------------------------

using System;
using System.IO;
using System.Collections;
using System.Threading;

namespace Bsc.WormBench
{
    /// <summary>
    /// Abstracts the BenchWorld - the underlying data structure that
    /// represents the matrix so that it can be changed.
    /// </summary>
    public class BenchWorld
    {
        #region Private Fields

        private string configFile;
        BenchWorldNode[,] world;
        int rowsNum;
        int columnsNum;

        #endregion

        #region Constructors

        public BenchWorld(string configFile)
        {
            this.configFile = configFile;
            Initialize();

            Config.Print(String.Format("BenchWorld {0}x{1}", rowsNum, columnsNum));
        }

        #endregion

        #region Public Methods

        public void DemoPrint()
        {
            if (!Config.IsDemo)
            {
                return;
            }

            lock (Config.ConsoleLock)
            {                
                ConsoleExtension.SetCursorPosition(0, 0);

                for (int i = 0; i < rowsNum; i++)
                {
                    for (int j = 0; j < columnsNum; j++)
                    {
                        if (world[i, j].WormGroupId == 0)
                        {
                            ConsoleExtension.ResetColor();
                        }
                        else
                        {
                            //
                            // 16 is the total number of available colors
                            //                            
                            ConsoleExtension.SetConsoleColor(
                                (ConsoleForeground)(((int)ConsoleForeground.Blue + world[i, j].WormGroupId) % 16),
                                ConsoleBackground.Black);
                        }
                        //Console.Write(world[i, j].WormGroupId + " ");
                        Console.Write(world[i, j].Value + " ");                        
                    }
                    Console.WriteLine();
                }
            }
        }

        //
        // Overloads the [,] operator.
        //
        public BenchWorldNode this[int x, int y]
        {
            get
            {
                x = x % rowsNum;
                y = y % rowsNum;

                if (x < 0)
                {
                    x = x + rowsNum;
                }

                if (y < 0)
                {
                    y = y + columnsNum;
                }

                return world[x, y];
            }
            set
            {
                x = x % rowsNum;
                y = y % rowsNum;

                if (x < 0)
                {
                    x = x + rowsNum;
                }

                if (y < 0)
                {
                    y = y + columnsNum;
                }

                world[x, y] = value;
            }
        }

        public BenchWorldNode this[Coordinate coordinate]
        {
            get
            {
                return this[coordinate.x, coordinate.y];
            }

            set
            {
                this[coordinate.x, coordinate.y] = value;
            }
        }

        #endregion

        #region Private Methods

        private void Initialize()
        {
            ParseConfigFile();
        }

        /// <summary>
        /// Assumes that the config file contains information for initializing a matrix
        /// where numbers are separated by space(s).
        /// </summary>
        /// <example>
        /// 3x3
        /// 1 2 3
        /// 4 5 6
        /// 7 8 9
        /// </example>
        private void ParseConfigFile()
        {
            using (StreamReader sr = new StreamReader(configFile))
            {
                if (sr.Peek() < 0)
                {
                    throw new ParseException("ROWSxCOLLUMNS expected", configFile, 1);
                }
                else 
                {
                    string line = sr.ReadLine();
                    line = line.Trim();

                    if (line != null && line.Length > 0 && line[0] != '#')
                    {
                        string[] tokens = line.Split('x');

                        try
                        {
                            rowsNum = Int32.Parse(tokens[0]);
                            columnsNum = Int32.Parse(tokens[1]);

                            world = new BenchWorldNode[rowsNum, columnsNum];
                        }
                        catch (FormatException e)
                        {
                            throw new ParseException(e.Message, configFile, 1);
                        }
                    }
                }

                int index = 0;

                try
                {
                    string line;
                    string[] tokens;

                    for (index = 0; index < rowsNum; index++)
                    {
                        line = sr.ReadLine();
                        tokens = line.Split(' ');

                        int colCounter = 0;
                        int colIndex = 0;

                        while ((colCounter < columnsNum) && (colIndex < tokens.Length))
                        {
                            if (tokens[colIndex] != null && tokens[colIndex].Length > 0)
                            {
                                world[index, colCounter] = new BenchWorldNode(Int32.Parse(tokens[colIndex]));
                                colCounter++;
                            }

                            colIndex++;
                        }

                        if (colCounter < columnsNum)
                        {
                            throw new ParseException("Not enough matrix elements in row.", configFile, index + 2);
                        }
                    }
                }
                catch (IOException e)
                {
                    throw new ParseException(e.Message, configFile, index + 2);
                }                
            }
        }

        #endregion

        #region Properties
        public int RowsNum
        {
            get
            {
                return rowsNum;
            }
        }

        public int ColumnsNum
        {
            get
            {
                return columnsNum;
            }
        }
        #endregion
    }
}
