/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: Config.cs
// Description: Application wide configuratin parameters.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 22.01.2008
//
//-----------------------------------------------------------------------------

//#define STM

using System;
using System.IO;
using System.Threading;

namespace Bsc.WormBench
{
    /// <summary>
    /// Application wide configuratin parameters.
    /// </summary>
    public class Config
    {
        #region Private Fields

        private static int threadsNum = 0;
        private static Worm[] worms;
        private static BenchWorld benchWorld;
        private static string wormsConfigFile;
        private static string benchWorldConfigFile;

        private static string[] configFileBuffer;

        private static string operationsConfigFile;
        private static Operation[] operations;

        private static bool isDemo;
        private static bool isVerify;
        private static bool isVerbose;

        private static int iterations;

        public static Object ConsoleLock;

#if LOCK
        public static Object GlobalLock;
#endif

        #endregion

        #region Public Fields
        //
        // We do not synchornize access to this assuming that concurrent update followed by
        // read is not semanitaclly wrong for us.
        // This field is public because to reduce any overhead of access through wrappers.
        //
        public static bool IsTimedOut;

        #endregion

        #region Public Methods

        /// <summary>
        /// Initialize the application wide configuration variables such as
        /// the worms, and the BenchWorld.
        /// </summary>
        /// <param name="threadsNum">
        /// The number of threads.
        /// </param>
        /// <param name="wormsConfigDir">
        /// The name of the directory containing the worms' configuration files.
        /// </param>
        /// <param name="benchWorldConfigFile">
        /// The name of the file that contains the configuration of the
        /// BenchWorld.
        /// </param>
        /// <param name="operationsConfigFile">
        /// The operations file.
        /// </param>
        /// <param name="isDemo">
        /// Specifis if this is a demo run.
        /// </param>
        /// <param name="isVerify">
        /// Specify if the results will be verified.
        /// </param>
        /// <param name="logFile">
        /// Log file.
        /// </param>
        /// <param name="iterations">
        /// How many times to repeate/iterate the execution of the stream
        /// of operations.
        /// </param>
        /// <exception cref="">Throws all IO exceptions.</exception>
        public static void Initialize(
            int threadsNum,
            string wormsConfigDir,
            string benchWorldConfigFile,
            string operationsConfigFile,
            bool isDemo, 
            bool isVerify,
            string logFile,
            int iterations)
        {            

#if STM
            Print("Synchronization: STM");
#elif LOCK
            Print("Synchronization: LOCK");
            GlobalLock = new Object();
#else // NULL_SYNCH
            Print("Synchronization: NULL");
#endif

            Config.threadsNum = threadsNum;
            Config.wormsConfigFile = wormsConfigDir;
            Config.benchWorldConfigFile = benchWorldConfigFile;
            Config.operationsConfigFile = operationsConfigFile;
            Config.iterations = iterations;

            Log.Initialize(logFile);            

            Config.isVerify = isVerify;
            Config.isVerbose = false;
            Config.IsTimedOut = false;

            InitializeBenchWorld();

            Config.isDemo = isDemo;
            if (Config.isDemo)
            {
                ConsoleLock = new Object();
                InitializeDemo();                
            }

            //
            // Operations shouls be initialized before worms. There is dependency.
            //
            InitializeOperations();
            
            InitializeWorms();
        }

        public static void Print(string message)
        {
            if (isVerbose)
            {
                Console.WriteLine(message);
            }
        }

        #endregion

        #region Private Methods

        private static void InitializeOperations()
        {
            operations = Operation.Load(Config.operationsConfigFile);
        }

        private static void InitializeBenchWorld()
        {            
            benchWorld = new BenchWorld(benchWorldConfigFile);
        }

        private static void ParseWormsConfigFile()
        {            
            using (StreamReader sr = new StreamReader(wormsConfigFile))
            {
                string buffer = sr.ReadToEnd();
                configFileBuffer = buffer.Split('\n');
            }

            for (int i = 0; i < configFileBuffer.Length; i++)
            {
                configFileBuffer[i] = configFileBuffer[i].Trim();
                configFileBuffer[i] = configFileBuffer[i].ToLower();
                if (configFileBuffer[i] != null && configFileBuffer[i].Length > 0 && configFileBuffer[i][0] == '#')
                {
                    configFileBuffer[i] = string.Empty;
                }
            }
        }

        private static void InitializeWorms()
        {
            ParseWormsConfigFile();            

            //
            // Check if there are defined enough worms in the file.
            //
            int wormsCount = 0;
            try
            {
                wormsCount = Int32.Parse(GetConfigValue("WormsCount"));                
            }
            catch (FormatException e)
            {
                throw new ParseException(e.Message, wormsConfigFile);
            }

            if (wormsCount < threadsNum)
            {
                throw new ParseException("There are less worm definitions that requested threads", wormsConfigFile);
            }

            int index = 0;
            worms = new Worm[threadsNum];
            int operationsChunkSize = operations.Length / threadsNum;
            for (int i = 0; i < configFileBuffer.Length; i++)
            {
                //
                // We distinguish the lines for the worm by checking if the line contains "wormid".
                // Of course, this is not precise.
                //
                if (configFileBuffer[i].IndexOf("wormid") != -1)
                {
                    if (index == threadsNum)
                    {
                        //
                        // Rturn when we read enough worm configurations - equal to the
                        // number of the threads.
                        //
                        break;                        
                    }

                    worms[index] = new Worm(configFileBuffer[i], index * operationsChunkSize, operationsChunkSize);
                    index++;
                }
            } 
        }

        private static string GetConfigValue(string configParamName)
        {
            string value = string.Empty;

            configParamName = configParamName.ToLower();

            bool found = false;
            int lineNo = 0;

            while ((!found) && (lineNo < configFileBuffer.Length))
            {
                if (configFileBuffer[lineNo].IndexOf(configParamName) != -1)
                {
                    string[] tokens = configFileBuffer[lineNo].Split('=');

                    if (tokens.Length == 2)
                    {
                        found = true;
                        value = tokens[1].Trim();
                    }
                }

                lineNo++;
            }

            return value;
        }

        private static void InitializeDemo()
        {
            //
            // Not supported by .NET 1.1
            //
            ConsoleExtension.Title = "BSC - WormBench - Demo";
            ConsoleExtension.Clear();
            BenchWorld.DemoPrint();
            //ConsoleExtension.SetBufferSize(80, 50);
            //ConsoleExtension.SetWindowSize(80, 50);            
        }        

        #endregion


        #region Properties

        public static Worm[] Worms
        {
            get { return Config.worms; }
            set { Config.worms = value; }
        }

        public static BenchWorld BenchWorld
        {
            get { return Config.benchWorld; }
            set { Config.benchWorld = value; }
        }

        public static string WormsConfigFile
        {
            get { return Config.wormsConfigFile; }
            set { Config.wormsConfigFile = value; }
        }

        public static bool IsDemo
        {
            get { return Config.isDemo; }
        }

        public static bool IsVerify
        {
            get { return Config.isVerify; }
        }

        /// <summary>
        /// The total number of operations executed by all worms together.
        /// </summary>
        public static long TotalWorkLoad
        {
            get
            {
                return operations.Length;
            }
        }

        public static Operation[] Operations
        {
            get
            {
                return operations;
            }
        }

        public static int Iterations
        {
            get
            {
                return iterations;
            }
        }

        #endregion
    }
}
