/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: Log.cs
// Description: Logging facility.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 22.01.2008
//
//-----------------------------------------------------------------------------

using System;
using System.IO;

namespace Bsc.WormBench
{
    /// <summary>
    /// Logging facility.
    /// </summary>
    public class Log
    {
        private static string logFile;        

        public static void Initialize(string logFile)
        {
            Log.logFile = logFile;            
        }

        public static void LogInfo(string message)
        {
            if (Log.logFile != null && Log.logFile.Length > 0)
            {
                string header = string.Empty;
                if (!File.Exists(Log.logFile))
                {
                    //
                    // [Worms#] [Throughut] [TotalWorkLoad] [ParallelTimeCycles] [ParallelTime] [TotalTimeCycles] [TotalTime] [SynchType] [NxM] [W-Body] [W-Head] [W-BW-Init]
                    //
                    //header += "-------------------------------------------------------------------------------------------------------------------------" + Environment.NewLine;
                    header = "Worms#,Throughut,TotalWorkLoad,ParallelTimeCycles,ParallelTime(sec),TotalTimeCycles(cpu),TotalTime(sec),SynchType,NxM,BodyLength,HeadSize,BwInit" + Environment.NewLine;
                }
 
                using (StreamWriter sw = File.AppendText(Log.logFile))
                {                    
                    sw.WriteLine(header + message);
                }
            }
            else
            {
                Console.WriteLine(message);
            }
        }

        public static void LogExecutionTime(long totalTime, long parallelTime)
        {
            int threadsNum = Config.Worms.Length;

            //
            // The total number of operations executed by all worms.
            // This may differ because of the timeout.
            //
            long totalWorkLoad = 0;
            for (int i = 0; i < Config.Worms.Length; i++)
            {
                totalWorkLoad += Config.Worms[i].ExecutedOperationsCount;
            }          

            //
            // Throughput per second
            //
            int throughPut = (int)(totalWorkLoad * 10000 / parallelTime);

#if STM
            string syncType = "ATOMIC";
#elif LOCK
            string syncType = "LOCK";
#else
            string syncType = "NULL";
#endif

            int n = Config.BenchWorld.RowsNum;
            int m = Config.BenchWorld.ColumnsNum;

            TimeSpan totalTimeWall = new TimeSpan(totalTime);
            TimeSpan parallelTimeWall = new TimeSpan(parallelTime);

            string[] wormConfigFileNameTokens = Config.WormsConfigFile.Split('-');
            string wormsBodyLength = "";
            string wormsHeadSize = "";
            string wormsBwInit = "";

            if (wormConfigFileNameTokens.Length > 1)
            {
                wormsBodyLength = wormConfigFileNameTokens[1];
            }

            if (wormConfigFileNameTokens.Length > 2)
            {
                wormsHeadSize = wormConfigFileNameTokens[2];
            }

            if (wormConfigFileNameTokens.Length > 3)
            {
                wormsBwInit = wormConfigFileNameTokens[3];
            }

            //
            // [Worms#] [Throughut] [TotalWorkLoad] [ParallelTimeCycles] [ParallelTime] [TotalTimeCycles] [TotalTime] [SynchType] [NxM] [W-Body] [W-Head] [W-BW-Init]
            //
            string logMessage = String.Format("{0},{1},{2},{3},{4},{5},{6},{7},{8},{9},{10},{11}",
                threadsNum,
                throughPut,
                totalWorkLoad,
                parallelTime,
                parallelTimeWall,
                totalTime,
                totalTimeWall,
                syncType,
                n,
                wormsBodyLength,
                wormsHeadSize,
                wormsBwInit);

            LogInfo(logMessage);
        }
    }
}
