/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: Operation.cs
// Description: Abstracts an operation intended to be executed by a worm.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 25.02.2008
//
//-----------------------------------------------------------------------------

using System;
using System.IO;

namespace Bsc.WormBench
{        
    /// <summary>
    /// Different operation types.
    /// </summary>
    public enum OperationType
    {
        SUM,                     // 0
        AVERAGE,                 // 1
        MEDIAN,                  // 2
        MIN,                     // 3
        MAX,                     // 4
        REPLACE_MAX_WITH_AVG,    // 5
        REPLACE_MIN_WITH_AVG,    // 6
        REPLACE_MEDIAN_WITH_AVG, // 7
        REPLACE_MAX_AND_MIN,     // 8
        REPLACE_MEDIAN_WITH_MIN, // 9
        REPLACE_MEDIAN_WITH_MAX, // 10
        SORT,                    // 11
        TRANSPOSE,               // 12
        CHECKPOINT,              // 13
        UNDO,                    // 14
        LEAVE_MESSAGE            // 15
    }

    /// <summary>
    /// Abstracts an operation intended to be executed by a worm.
    /// </summary>
    public class Operation
    {
        #region Private Fields
        private OperationType operationType;
        private Direction direction;
        #endregion

        #region Constructors
        public Operation()
        {
            this.operationType = OperationType.SUM;
            this.direction = Direction.Ahead;
        }

        public Operation(OperationType operationType, Direction direction)
        {
            this.operationType = operationType;
            this.direction = direction;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Load the operations from a file. The format of the file should be:
        /// 1. Number Of Operations
        /// 2. OperationType - Direction
        /// 3. OperationType - Direction
        /// ...
        /// </summary>
        /// <param name="configFile">The file to load the operations from.</param>
        /// <returns>An array of the operations.</returns>
        public static Operation[] Load(string configFile)
        {
            using (StreamReader sr = new StreamReader(configFile))
            {
                string numberOfOperationsStr = sr.ReadLine();
                long numberOfOperations = Int64.Parse(numberOfOperationsStr);

                Operation[] operations = new Operation[numberOfOperations];

                string line;
                string[] tokens;                

                for (long i = 0; i < numberOfOperations; i++)
                {
                    line = sr.ReadLine();
                    line = line.Trim();

                    tokens = line.Split('-');

                    if (tokens == null || tokens.Length != 2)
                    {
                        throw new ParseException(
                            "Parsing operations. Expected OperationType - Cidrection", 
                            configFile,
                            (int)i + 1);
                    }

                    tokens[0] = tokens[0].Trim();
                    tokens[1] = tokens[1].Trim();

                    try
                    {
                        operations[i] = new Operation(
                            (OperationType)Int32.Parse(tokens[0]),
                            (Direction)Int32.Parse(tokens[1]));
                    }
                    catch (FormatException e)
                    {
                        throw new ParseException(e.Message, configFile, (int)i + 1);
                    }
                }

                return operations;
            }            
        }

        #endregion

        #region Private Utility Methods        
        #endregion

        #region Properties
        public OperationType OperationType
        {
            get
            {
                return this.operationType;
            }
        }

        public Direction Direction
        {
            get
            {
                return this.direction;
            }
        }
        #endregion
    }
}
