/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: Worm.cs
// Description: Abstracts the worm object. This is an active object.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 22.01.2008
//
//-----------------------------------------------------------------------------

//#define STM

using System;
using System.Threading;
using System.Collections;

namespace Bsc.WormBench
{
    public enum Direction
    {
        Ahead,
        Right,
        Left
    }

    public struct Coordinate
    {
        public int x;
        public int y;
    }

    /// <summary>
    /// Abstracts the worm object. This is an active object.
    /// </summary>
    public class Worm
    {
        #region Private Fields

        private string configStr;
        private string[] configStrTokens;

        private int id;
        private int groupId;
        private string name;
        private int headSize;
        private int speed;
        private int operationsStartIndex;
        private int operationsNum;

        //
        // The case when bodyLength != body.Length is when we want a worm with
        // body of length 1. In this case we again have body[0] and body[1] cells
        // but use body[1] for finding the orientation of the worm. This does not
        // have any influence except writing the cells under the body - then
        // we don't update the cell corresponding to body[1].
        //
        private int bodyLength;
        private Coordinate[] body;
        private Coordinate[] expectedBody;

        //
        // The mapping between the headBuffer indices and the original coordinates
        // in BenchWorld. This is updated when buffer is being read.
        //
        private Coordinate[] headCoordinates;
        private int[] headBuffer;        

        private Queue resultsBuffer;
        private int currentOperationIndex;

        private int operationsCounter;

        Thread thread;

        #endregion

        #region Constructors

        public Worm(string configStr, int operationsStartIndex, int operationsNum)
        {
            this.configStr = configStr;
            this.operationsStartIndex = operationsStartIndex;
            this.operationsNum = operationsNum;
            this.currentOperationIndex = 0;

            resultsBuffer = new Queue(100);

            thread = new Thread(new ThreadStart(Run));

            Initialize();

            Config.Print(name + " initialized");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Start the worm.
        /// </summary>
        public void Start()
        {            
            thread.Start();
        }

        public void Join()
        {
            thread.Join();
        }

        public void Join(int millisecondsTimeout)
        {
            thread.Join(millisecondsTimeout);
        }

        public void Join(TimeSpan timeToWait)
        {
            thread.Join(timeToWait);
        }

        /// <summary>
        /// The entry point of the Worm thread.
        /// </summary>
        public void Run()
        {
            //ThreadHelper.SetThreadAffinityToCpu(Thread.CurrentThread, (uint)((id % Config.Worms.Length) + 1));

            //
            // Execute the same set of operations 10 times repeatedtly.
            // The reason for this is that loading of the file take too much time
            // that should be somehow compensated.
            //
            this.operationsCounter = this.operationsStartIndex;

            while ((operationsCounter - operationsStartIndex < operationsNum * Config.Iterations) && (!Config.IsTimedOut))
            {
                currentOperationIndex = operationsCounter % Config.Operations.Length;
                Operation currentOperation = Config.Operations[currentOperationIndex];

                ReadNodesUnderHead();
                ExecuteOperation(currentOperation.OperationType);

                Move(currentOperation.Direction);

                operationsCounter++;                
            }
            
            Verify();
        }

        #endregion

        #region Private Methods        

        /// <summary>
        /// Read the nodes under the head and buffer them.
        /// The operation of reading all the cells under the head is atomic.
        /// </summary>
        private void ReadNodesUnderHead()
        {
            ErasePrintedHead();

            Coordinate diff;

            diff.x = body[0].x - body[1].x;
            diff.y = body[0].y - body[1].y;           
            
            int headIndex = 0;
#if STM            
            try
#elif LOCK
            lock(Config.GlobalLock)
#endif
            {
                for (int i = 0; i <= headSize; i++)
                {
                    for (int j = -headSize + i + 1; j <= headSize - i - 1; j++)
                    {
                        int x = 0;
                        int y = 0;

                        if (diff.x == 1 && diff.y == 0)
                        {
                            //
                            // Up.
                            //
                            x = body[0].x - i;
                            y = body[0].y + j;
                        }
                        else if (diff.x == -1 && diff.y == 0)
                        {
                            //
                            // Down.
                            //
                            x = body[0].x + i;
                            y = body[0].y + j; 
                        }
                        else if (diff.x == 0 && diff.y == 1)
                        {
                            //
                            // Right.
                            //
                            x = body[0].x + j;
                            y = body[0].y + i;

                        }
                        else if (diff.x == 0 && diff.y == -1)
                        {
                            //
                            // Left.
                            //
                            x = body[0].x + j;
                            y = body[0].y - i;
                        }

                        headCoordinates[headIndex].x = x;
                        headCoordinates[headIndex].y = y;
                        headBuffer[headIndex] = Config.BenchWorld[x, y].Value;
                        headIndex++;
                    }
                }
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        /// <summary>
        /// Write the nodes under the worm's head.
        /// </summary>
        private void WriteNodesUnderHead()
        {
#if STM
            try
#elif LOCK
            lock(Config.GlobalLock)
#endif
            {
                for (int headIndex = 0; headIndex < headBuffer.Length; headIndex++)
                {
                    Coordinate benchWorldCoordinate = HeadIndexToBenchWorldCoordinate(headIndex);
                    Config.BenchWorld[benchWorldCoordinate].Value = headBuffer[headIndex];
                }
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        /// <summary>
        /// Performs a computation/operation on the buffered nodes under
        /// the head and stores them in a results buffer/queue.
        /// This operation may be atomic 
        /// </summary>
        /// <param name="command">
        /// The command to execute.
        /// </param>
        private void ExecuteOperation(OperationType operation)
        {            
            switch (operation)
            {
                case OperationType.SUM: Sum(); break;
                case OperationType.AVERAGE: Average(); break;
                case OperationType.MAX: Maximum(); break;
                case OperationType.MEDIAN: Median(); break;
                case OperationType.MIN: Minimum(); break;
                case OperationType.REPLACE_MAX_AND_MIN: ReplaceMaxAndMin(); break;
                case OperationType.REPLACE_MAX_WITH_AVG: ReplaceMaxWithAverage(); break;
                case OperationType.REPLACE_MEDIAN_WITH_AVG: ReplaceMedianWithAverage(); break;
                case OperationType.REPLACE_MEDIAN_WITH_MAX: ReplaceMedianWithMax(); break;
                case OperationType.REPLACE_MEDIAN_WITH_MIN: ReplaceMedianWithMin(); break;
                case OperationType.REPLACE_MIN_WITH_AVG: ReplaceMinWithAverage(); break;
                case OperationType.TRANSPOSE: Transpose(); break;
                case OperationType.SORT: Sort(); break;
            }
        }

        /// <summary>
        /// Move the body of the worm. An atomic operation.
        /// </summary>
        /// <param name="direction">
        /// The direction to move to.
        /// </param>
        /// <see cref="Direction"/>
        private void Move(Direction direction)
        {
            Coordinate diff;
            Coordinate newCoordinate;

            newCoordinate.x = 0;
            newCoordinate.y = 0;
            
            diff.x = body[0].x - body[1].x;
            diff.y = body[0].y - body[1].y;

            switch (direction)
            {
                case Direction.Ahead:
                    {
                        newCoordinate.x = Math.Abs(diff.x) * (body[0].x + speed * diff.x) + Math.Abs(diff.y) * body[0].x;
                        newCoordinate.y = Math.Abs(diff.y) * (body[0].y + speed * diff.y) + Math.Abs(diff.x) * body[0].y;
                    } break;
                case Direction.Right:
                    {
                        newCoordinate.x = Math.Abs(diff.y) * (body[0].x - speed * diff.y) + Math.Abs(diff.x) * body[0].x;
                        newCoordinate.y = Math.Abs(diff.x) * (body[0].y - speed * diff.x) + Math.Abs(diff.y) * body[0].y;
                    } break;
                case Direction.Left:
                    {
                        newCoordinate.x = Math.Abs(diff.y) * (body[0].x + speed * diff.y) + Math.Abs(diff.x) * body[0].x;
                        newCoordinate.y = Math.Abs(diff.x) * (body[0].y + speed * diff.x) + Math.Abs(diff.y) * body[0].y;
                    } break;
            }

            //
            // Changing the coordinates of the Worm is necessary to be atomic because
            // another Worm may change its coordinates concurently as its part of
            // execution of the operations.
            //
#if STM            
            try
#elif LOCK
            lock(Config.GlobalLock)
#endif
            {
                Coordinate oldCoordinate;
                for (int i = 0; i < body.Length; i++)
                {
                    oldCoordinate = body[i];
                    body[i] = newCoordinate;
                    newCoordinate = oldCoordinate;
                }
            }
#if STM
            catch (AtomicFakeException) {}            
#endif
            
            //
            // Write on the cells below the body. This simulates the case of barriear.
            // Worms from the same group can cross each other.
            //
#if STM            
            try
#elif LOCK
            lock(Config.GlobalLock)
#endif
            {
                for (int i = 0; i < bodyLength; i++)
                {
                    Config.BenchWorld[body[i]].WormGroupId = groupId;
                }
            }
#if STM
            catch (AtomicFakeException) {}
#endif
            
            if (bodyLength != 1)
            {
                //
                // Actually at this point newCoordinate is the old coordinate.
                //
                Config.BenchWorld[newCoordinate].WormGroupId = 0;
                DemoPrint(newCoordinate);
            }
            else
            {
                Config.BenchWorld[body[1]].WormGroupId = 0;
                DemoPrint(body[1]);
            }
            
            //Config.BenchWorld.DemoPrint();
        }

        /// <summary>
        /// Computes the direction by sum(body)%3.
        /// </summary>
        /// <returns>The direction to go to.</returns>
        /// <remarks>This is atomic.</remarks>
        private Direction ComputeDirection()
        {
            return Direction.Ahead;
        }

        private void Initialize()
        {
            ParseConfigStr();

            string strId = GetConfigValue("WormID");
            if (strId != null && strId.Length > 0)
            {
                id = Int32.Parse(strId);
            }
            else
            {
                throw new ParseException("Worm ID not defined", configStr); 
            }

            string strGroupId = GetConfigValue("GroupID");
            if (strGroupId != null && strGroupId.Length > 0)
            {
                groupId = Int32.Parse(strGroupId);
            }
            else
            {
                throw new ParseException("Worm GroupID not defined", configStr);
            }

            name = GetConfigValue("Name");

            string strHeadSize = GetConfigValue("HeadSize");
            if (strHeadSize != null && strHeadSize.Length > 0)
            {
                headSize = Int32.Parse(strHeadSize);

                //
                // n*n
                //
                headBuffer = new int[headSize * headSize];
                headCoordinates = new Coordinate[headSize * headSize];
            }
            else
            {
                throw new ParseException("Worm HeadSize not defined", configStr);
            }

            string strSpeed = GetConfigValue("Speed");
            if (strSpeed != null && strSpeed.Length > 0)
            {
                speed = Int32.Parse(strSpeed);
            }
            else
            {
                throw new ParseException("Worm Speed not defined", configStr);
            }

            string bodyLengthStr = GetConfigValue("BodyLength");
            if (bodyLengthStr != null && bodyLengthStr.Length > 0)
            {
                bodyLength = Int32.Parse(bodyLengthStr);
            }
            else
            {
                throw new ParseException("Worm BodyLength not defined", configStr);
            }

            string strBody = GetConfigValue("Body");
            if (strBody != null && strBody.Length > 0)
            {
                string[] tokens = strBody.Split(']');

                if (bodyLength == 1)
                {
                    //
                    // If the body lenght is 1 we yet create one more cell
                    // that we later use to compute the orientation of the worm.
                    //
                    body = new Coordinate[bodyLength + 1];
                }
                else
                {
                    body = new Coordinate[bodyLength];
                }

                if (tokens.Length  - 1 != body.Length)
                {
                    throw new ParseException("Worm BodyLength not given body coordinates does not match", configStr);
                }
                               
                for (int i = 0; i < body.Length; i++)
                {
                    tokens[i] = tokens[i].Replace("[", "");
                    string[] strCoordinate = tokens[i].Split(',');

                    body[i] = new Coordinate();
                    body[i].x = Int32.Parse(strCoordinate[0]);
                    body[i].y = Int32.Parse(strCoordinate[1]);
                }

                //
                // Copy the body to expectedBody so that we can verify the execution in the
                // Verify method.
                //
                expectedBody = new Coordinate[body.Length];
                Array.Copy(body, expectedBody, body.Length);
            }
            else
            {
                throw new ParseException("Worm Body not defined", configStr);
            }
        }

        private void ParseConfigStr()
        {
            //
            // The worm is defined as
            // WormID = 1; Name = Worm1; GroupID = 1; HeadSize = 2; Speed = 1; Body = [23,1][23,2][23,3][24,3]
            // Where ";" is a delimeter.
            //
            configStrTokens = configStr.Split(';');

            for (int i = 0; i < configStrTokens.Length; i++)
            {
                configStrTokens[i] = configStrTokens[i].Trim();
                configStrTokens[i] = configStrTokens[i].ToLower();                
            }
        }

        private string GetConfigValue(string configParamName)
        {
            configParamName = configParamName.ToLower();
            string value = string.Empty;

            bool found = false;
            int lineNo = 0;

            while ((!found) && (lineNo < configStrTokens.Length))
            {
                if (configStrTokens[lineNo].IndexOf(configParamName) != -1)
                {
                    string[] tokens = configStrTokens[lineNo].Split('=');

                    if (tokens.Length == 2 && tokens[0].Trim() == configParamName)
                    {
                        found = true;
                        value = tokens[1].Trim();
                    }
                }

                lineNo++;
            }

            return value;
        }

        /// <summary>
        /// Converts an head index into an appropriate coordinate into the matrix.
        /// </summary>
        /// <param name="headIndex">
        /// The head index.
        /// </param>
        /// <returns>
        /// A repsective mapping of the head index into the matrix.
        /// </returns>
        private Coordinate HeadIndexToBenchWorldCoordinate(int headIndex)
        {
            return headCoordinates[headIndex];
        }

        /// <summary>
        /// Verifies that the execution is correct.
        /// The verification is done by simulating all the movements of the worm
        /// and then comparing the actual coordinates of the body with the expected.
        /// 
        /// Further verification can be done by checking the results buffered by the
        /// worm during its execution.
        /// </summary>
        /// <remarks>
        /// The current implementation of this method is exatcly the same like Move,
        /// just the BenchWorld is not updated.
        /// </remarks>
        private void Verify()
        {
            if (!Config.IsVerify)
            {
                return;
            }

            // TODO: fix this method
            int operationsCounter = operationsStartIndex;

            while (operationsCounter - operationsStartIndex < operationsNum * Config.Iterations)
            {
                currentOperationIndex = operationsCounter % Config.Operations.Length;
                Operation currentOperation = Config.Operations[currentOperationIndex];

                Coordinate diff;
                Coordinate newCoordinate;

                newCoordinate.x = 0;
                newCoordinate.y = 0;

                diff.x = expectedBody[0].x - expectedBody[1].x;
                diff.y = expectedBody[0].y - expectedBody[1].y;

                switch (currentOperation.Direction)
                {
                    case Direction.Ahead:
                        {
                            newCoordinate.x = Math.Abs(diff.x) * (expectedBody[0].x + speed * diff.x) + Math.Abs(diff.y) * expectedBody[0].x;
                            newCoordinate.y = Math.Abs(diff.y) * (expectedBody[0].y + speed * diff.y) + Math.Abs(diff.x) * expectedBody[0].y;
                        } break;
                    case Direction.Right:
                        {
                            newCoordinate.x = Math.Abs(diff.y) * (expectedBody[0].x - speed * diff.y) + Math.Abs(diff.x) * expectedBody[0].x;
                            newCoordinate.y = Math.Abs(diff.x) * (expectedBody[0].y - speed * diff.x) + Math.Abs(diff.y) * expectedBody[0].y;
                        } break;
                    case Direction.Left:
                        {
                            newCoordinate.x = Math.Abs(diff.y) * (expectedBody[0].x + speed * diff.y) + Math.Abs(diff.x) * expectedBody[0].x;
                            newCoordinate.y = Math.Abs(diff.x) * (expectedBody[0].y + speed * diff.x) + Math.Abs(diff.y) * expectedBody[0].y;
                        } break;
                }

                //
                // This code actually implements the move method.
                // But we need not to alter the original body of the worm so we do it in
                // isolation.
                //
                Coordinate oldCoordinate;
                for (int i = 0; i < expectedBody.Length; i++)
                {
                    oldCoordinate = expectedBody[i];
                    expectedBody[i] = newCoordinate;
                    newCoordinate = oldCoordinate;
                }

                operationsCounter++;
            }

            for (int i = 0; i < body.Length; i++)
            {
                if (body[i].x != expectedBody[i].x || body[i].y != expectedBody[i].y)
                {                    
                    Console.WriteLine(
                        this.name + ": Verification failed. Expected [{0},{1}], but was [{2}, {3}]",
                        expectedBody[i].x,
                        expectedBody[i].y,
                        body[i].x,
                        body[i].y);

                    return;
                }
            }

            Config.Print(this.name + ": Verification successful.");
        }

        private void DemoPrint(Coordinate erasedNode)
        {
            if (!Config.IsDemo)
            {
                return;
            }

            lock (Config.ConsoleLock)
            {
                ConsoleExtension.SetCursorPosition(erasedNode.y * 2, erasedNode.x);
                Console.Write(Config.BenchWorld[erasedNode].Value + " ");

                ConsoleExtension.SetConsoleColor(
                                (ConsoleForeground)(((int)ConsoleForeground.Blue + groupId) % 16),
                                ConsoleBackground.Black);

                //
                // Print the body
                //
                for (int i = 0; i < bodyLength; i++)
                {
                    ConsoleExtension.SetCursorPosition(body[i].y * 2, body[i].x);
                    Console.Write(Config.BenchWorld[body[i]].Value + " ");
                }

                //
                // Print the head
                //
                for (int i = 0; i < headCoordinates.Length; i++)
                {
                    ConsoleExtension.SetCursorPosition(headCoordinates[i].y * 2, headCoordinates[i].x);
                    Console.Write(Config.BenchWorld[headCoordinates[i]].Value + " ");
                }

                ConsoleExtension.ResetColor();                
            }
        }

        private void ErasePrintedHead()
        {
            if (!Config.IsDemo)
            {
                return;
            }

            //
            // Just reprint with the normal colors.
            //
            lock (Config.ConsoleLock)
            {
                for (int i = 0; i < headCoordinates.Length; i++)
                {
                    ConsoleExtension.SetCursorPosition(headCoordinates[i].y * 2, headCoordinates[i].x);
                    Console.Write(Config.BenchWorld[headCoordinates[i]].Value + " ");
                }
            }
        }

        #endregion

        #region Worm Operations

        /// <summary>
        /// Sum the cells below the head.
        /// </summary>
        private void Sum()
        {       
            int sum = 0;

            for (int i = 0; i < headBuffer.Length; i++)
            {
                sum += headBuffer[i];
            }

            if (resultsBuffer.Count == Definitions.RESULTS_QUEUE_LIMIT)
            {
                resultsBuffer.Dequeue();
            }
            
            resultsBuffer.Enqueue(sum);
        }

        /// <summary>
        /// Find the average of Worm's head.
        /// </summary>
        private void Average()
        {
            int sum = 0;

            for (int i = 0; i < headBuffer.Length; i++)
            {
                sum += headBuffer[i];
            }

            int average = sum / headBuffer.Length;

            if (resultsBuffer.Count == Definitions.RESULTS_QUEUE_LIMIT)
            {
                resultsBuffer.Dequeue();
            }

            resultsBuffer.Enqueue(average);
        }

        private void Median()
        {
            //
            // Selection sort. Intentionally choosen assumning that one would use less
            // efficient sorting algorithm.
            //
            for (int i = 0; i < headBuffer.Length; i++)
            {
                int min = headBuffer[i];
                int minIndex = i;

                for (int j = i + 1; j < headBuffer.Length; j++)
                {
                    if (headBuffer[j] < min)
                    {
                        min = headBuffer[j];
                        minIndex = j;
                    }
                }

                headBuffer[minIndex] = headBuffer[i];
                headBuffer[i] = min;
            }

            int median = headBuffer[headBuffer.Length / 2];

            if (resultsBuffer.Count == Definitions.RESULTS_QUEUE_LIMIT)
            {
                resultsBuffer.Dequeue();
            }

            resultsBuffer.Enqueue(median);
        }

        private void Minimum()
        {
            int min = headBuffer[0];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] < min)
                {
                    min = headBuffer[i];
                }
            }

            if (resultsBuffer.Count == Definitions.RESULTS_QUEUE_LIMIT)
            {
                resultsBuffer.Dequeue();
            }

            resultsBuffer.Enqueue(min);
        }

        private void Maximum()
        {
            int max = headBuffer[0];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] > max)
                {
                    max = headBuffer[i];
                }
            }

            if (resultsBuffer.Count == Definitions.RESULTS_QUEUE_LIMIT)
            {
                resultsBuffer.Dequeue();
            }

            resultsBuffer.Enqueue(max);
        }

        private void ReplaceMaxWithAverage()
        {
            int maxIndex = 0;
            int max = headBuffer[maxIndex];

            int sum = headBuffer[0];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] > max)
                {                    
                    maxIndex = i;
                    max = headBuffer[maxIndex];
                }

                sum += headBuffer[i];
            }

            int average = sum / headBuffer.Length;
            
            Coordinate maxCoordinate = HeadIndexToBenchWorldCoordinate(maxIndex);

#if STM
            try
#elif LOCK
            lock (Config.GlobalLock)
#endif
            {
                Config.BenchWorld[maxCoordinate].Value = average;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void ReplaceMinWithAverage()
        {
            int minIndex = 0;
            int min = headBuffer[minIndex];

            int sum = headBuffer[0];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] < min)
                {
                    minIndex = i;
                    min = headBuffer[minIndex];
                }

                sum += headBuffer[i];
            }

            int average = sum / headBuffer.Length;

            Coordinate minCoordinate = HeadIndexToBenchWorldCoordinate(minIndex);

#if STM
            try
#elif LOCK
            lock (Config.GlobalLock)
#endif
            {
                Config.BenchWorld[minCoordinate].Value = average;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void ReplaceMedianWithAverage()
        {
            int[] originalIndices = new int[headBuffer.Length];
            for (int i = 0; i < headBuffer.Length; i++)
            {
                originalIndices[i] = i;
            }

            //
            // Selection sort. Intentionally choosen assumning that one would use less
            // efficient sorting algorithm.
            //
            for (int i = 0; i < headBuffer.Length; i++)
            {
                int min = headBuffer[i];
                int minIndex = i;                

                for (int j = i + 1; j < headBuffer.Length; j++)
                {
                    if (headBuffer[j] < min)
                    {
                        min = headBuffer[j];
                        minIndex = j;
                    }
                }

                headBuffer[minIndex] = headBuffer[i];
                headBuffer[i] = min;

                //
                // Keep track of how the indices are mixed.
                //
                originalIndices[minIndex] = i;
                originalIndices[i] = minIndex;
            }

            int medianOriginalIndex = originalIndices[headBuffer.Length / 2];

            int sum = 0;
            for (int i = 0; i < headBuffer.Length; i++)
            {
                sum += headBuffer[i];
            }

            int average = sum / headBuffer.Length;            

            Coordinate medianIndexWormBenchCoordinate = HeadIndexToBenchWorldCoordinate(medianOriginalIndex);
#if STM
            try
#elif LOCK
            lock (Config.GlobalLock)
#endif
            {
                Config.BenchWorld[medianIndexWormBenchCoordinate].Value = average;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void ReplaceMaxAndMin()
        {
            int min = headBuffer[0];
            int minIndex = 0;

            int max = headBuffer[0];
            int maxIndex = 0;

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] < min)
                {
                    min = headBuffer[i];
                    minIndex = i;
                }

                if (headBuffer[i] > max)
                {
                    max = headBuffer[i];
                    maxIndex = i;
                }
            }

            Coordinate minCoordinate = HeadIndexToBenchWorldCoordinate(minIndex);
            Coordinate maxCoordinate = HeadIndexToBenchWorldCoordinate(maxIndex);

#if STM
            try
#elif LOCK
            lock(Config.GlobalLock)
#endif
            {
                Config.BenchWorld[minCoordinate].Value = max;
                Config.BenchWorld[maxCoordinate].Value = min;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void ReplaceMedianWithMin()
        {
            int minIndex = 0;
            int min = headBuffer[minIndex];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] < min)
                {
                    minIndex = i;
                    min = headBuffer[minIndex];
                }
            }

            Coordinate minCoordinate = HeadIndexToBenchWorldCoordinate(minIndex);

            int[] originalIndices = new int[headBuffer.Length];
            for (int i = 0; i < headBuffer.Length; i++)
            {
                originalIndices[i] = i;
            }

            //
            // Selection sort. Intentionally choosen assumning that one would use less
            // efficient sorting algorithm.
            //
            for (int i = 0; i < headBuffer.Length; i++)
            {
                int localMin = headBuffer[i];
                int localMinIndex = i;

                for (int j = i + 1; j < headBuffer.Length; j++)
                {
                    if (headBuffer[j] < localMin)
                    {
                        localMin = headBuffer[j];
                        localMinIndex = j;
                    }
                }

                headBuffer[localMinIndex] = headBuffer[i];
                headBuffer[i] = localMin;

                //
                // Keep track of how the indices are mixed.
                //
                originalIndices[localMinIndex] = i;
                originalIndices[i] = localMinIndex;
            }

            int medianValue = headBuffer[headBuffer.Length / 2];
            int medianOriginalIndex = originalIndices[headBuffer.Length / 2];

            Coordinate medianCoordinate = HeadIndexToBenchWorldCoordinate(medianOriginalIndex);

#if STM
            try
#elif LOCK
            lock (Config.GlobalLock)
#endif
            {
                Config.BenchWorld[minCoordinate].Value = medianValue;
                Config.BenchWorld[medianCoordinate].Value = min;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void ReplaceMedianWithMax()
        {
            int maxIndex = 0;
            int max = headBuffer[maxIndex];

            for (int i = 1; i < headBuffer.Length; i++)
            {
                if (headBuffer[i] < max)
                {
                    maxIndex = i;
                    max = headBuffer[maxIndex];
                }
            }

            Coordinate maxCoordinate = HeadIndexToBenchWorldCoordinate(maxIndex);

            int[] originalIndices = new int[headBuffer.Length];
            for (int i = 0; i < headBuffer.Length; i++)
            {
                originalIndices[i] = i;
            }

            //
            // Selection sort. Intentionally choosen assumning that one would use less
            // efficient sorting algorithm.
            //
            for (int i = 0; i < headBuffer.Length; i++)
            {
                int localMax = headBuffer[i];
                int localMaxIndex = i;

                for (int j = i + 1; j < headBuffer.Length; j++)
                {
                    if (headBuffer[j] > localMax)
                    {
                        localMax = headBuffer[j];
                        localMaxIndex = j;
                    }
                }

                headBuffer[localMaxIndex] = headBuffer[i];
                headBuffer[i] = localMax;

                //
                // Keep track of how the indices are mixed.
                //
                originalIndices[localMaxIndex] = i;
                originalIndices[i] = localMaxIndex;
            }

            int medianValue = headBuffer[headBuffer.Length / 2];
            int medianOriginalIndex = originalIndices[headBuffer.Length / 2];

            Coordinate medianCoordinate = HeadIndexToBenchWorldCoordinate(medianOriginalIndex);

#if STM
            try
#elif LOCK
            lock (Config.GlobalLock)
#endif
            {
                Config.BenchWorld[maxCoordinate].Value = medianValue;
                Config.BenchWorld[medianCoordinate].Value = max;
            }
#if STM
            catch (AtomicFakeException) { }
#endif
        }

        private void Sort()
        {
            //
            // Selection sort. Intentionally choosen assumning that one would use less
            // efficient sorting algorithm.
            //
            for (int i = 0; i < headBuffer.Length; i++)
            {
                int min = headBuffer[i];
                int minIndex = i;

                for (int j = i + 1; j < headBuffer.Length; j++)
                {
                    if (headBuffer[j] < min)
                    {
                        min = headBuffer[j];
                        minIndex = j;
                    }
                }

                headBuffer[minIndex] = headBuffer[i];
                headBuffer[i] = min;
            }

            WriteNodesUnderHead();
        }

        private void Transpose()
        {
            for (int i = 0; i < headBuffer.Length / 2; i++)
            {
                int temp = headBuffer[i];
                headBuffer[i] = headBuffer[headBuffer.Length - 1];
                headBuffer[headBuffer.Length - i - 1] = temp;
            }

            WriteNodesUnderHead();
        }

        private void Checkpoint()
        {
            throw new Exception("Worm::Checkpoint method not implemented");
        }

        private void Undo()
        {
            throw new Exception("Worm::Undo method not implemented");
        }

        private void LeaveMessage()
        {
            throw new Exception("Worm::LeaveMessage method not implemented");
        }

        #endregion

        #region Properties
        public int ExecutedOperationsCount
        {
            get
            {
                return this.operationsCounter - this.operationsStartIndex;
            }
        }
        #endregion
    }
}
