/*
WormBench - Synthetic workload for Transactinal Memory Systems
Copyright (C) 2008  Barcelona Supercomputing Center www.bscmsrc.eu

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//+----------------------------------------------------------------------------
//
// File Name: WormBench.cs
// Description: That main class of the WormBench application.
// Organization: Barcelona Supercomputing Center
// Author: Ferad Zyulkyarov ferad.zyulkyarov[@]bsc.es
// Date: 22.01.2008
//
// Supported Synchornizations:
// 1. LOCK - Coars grain lock
// 2. STM - Bartok STM
// 3. NULL_SYNCH - no synchornization, used for measuring throughput.
// 
// The support for the different synchronization is implemented by using
// preprocessor directives that check weather one of the above symbols are
// defined. The default is NULL_SYNCH. To specify the synchronization type
// the nmake variable SYNCHRONIZATION should be defined.
// For example: nmake SYNCHRONIZATION={LOCK, STM, NULL_SYNCH}
//
//-----------------------------------------------------------------------------

using System;
using System.Threading;

namespace Bsc.WormBench
{
    /// <summary>
    /// That main class of the WormBench application.
    /// </summary>
    public class WormBench
    {
        //
        // Timing
        //
        private static long totalTimeStart;
        private static long totalTimeStop;
        private static long parallelTimeStart;
        private static long parallelTimeStop;        

        //
        // Command line configuration arguments
        //
        private static int threadsNum = 0;
        private static string wormsDir = string.Empty;
        private static string benchWorldConfigFile = string.Empty;
        private static string operationsConfigFile = string.Empty;  
        private static bool isDemo = false;
        private static bool isVerify = false;
        private static string logFile = string.Empty;
        private static int iterations = 1;
        private static int timeoutSeconds = 0;

        public static void Main(string[] args)
        {
            totalTimeStart = System.DateTime.Now.Ticks;

            PrintHeader();

            if (!ParseArguments(args))
            {
                return;
            }

            Config.Initialize(
                threadsNum,
                wormsDir, 
                benchWorldConfigFile,
                operationsConfigFile, 
                isDemo, 
                isVerify, 
                logFile,
                iterations);

            RunBenchmark();
                        
            totalTimeStop = System.DateTime.Now.Ticks;
            Log.LogExecutionTime(totalTimeStop - totalTimeStart, parallelTimeStop - parallelTimeStart);
            PrintFooter();
        }

        private static bool ParseArguments(string[] args)
        {
            bool isSuccess = true;

            //
            // Should be args.Length > 10
            // This hack was done because of the automated tests
            //
            if (args.Length < 10 || args.Length > 14)
            {
                Console.WriteLine();
                Console.WriteLine("Error: Wrong command line arguments.");
                PrintUsage();                
                return false;
            }

            for (int i = 0; i < args.Length; i++)
            {
                if (args[i] == "-t")
                {
                    threadsNum = Int32.Parse(args[i + 1]);
                    CpuTools.CpusCount = (uint)threadsNum;
                }
                else if (args[i].IndexOf("-worms") != -1)
                {
                    wormsDir = args[i + 1];
                }
                else if (args[i].IndexOf("-benchworld") != -1)
                {
                    benchWorldConfigFile = args[i + 1];
                }
                else if (args[i].IndexOf("-operations") != -1)
                {
                    operationsConfigFile = args[i + 1].Trim();
                }
                else if (args[i].IndexOf("-demo") != -1)
                {
                    isDemo = true;
                }
                else if (args[i].IndexOf("-verify") != -1)
                {
                    isVerify = true;
                }
                else if (args[i].IndexOf("-log") != -1)
                {
                    logFile = args[i + 1].Trim();
                }
                else if (args[i].IndexOf("-iterations") != -1)
                {
                    iterations = Int32.Parse(args[i + 1]);
                }
                else if (args[i].IndexOf("-timeout") != -1)
                {
                    timeoutSeconds = Int32.Parse(args[i + 1]);
                }
            }

            if (threadsNum == 0)
            {
                Console.WriteLine("Error: -t not specified.");
                isSuccess = false;
            }

            if (wormsDir == string.Empty)
            {                
                Console.WriteLine("Error: -worms not specified.");
                isSuccess = false;
            }

            if (benchWorldConfigFile == string.Empty)
            {
                Console.WriteLine("Error: -benchworld not specified.");
                isSuccess = false;
            }

            if (operationsConfigFile == string.Empty)
            {
                Console.WriteLine("Error: -operations not specified.");
                isSuccess = false;                
            }

            if (!isSuccess)
            {
                Console.WriteLine();
                PrintUsage();
            }

            return isSuccess;
        }

        private static void PrintHeader()
        {
            Console.WriteLine();
            Console.WriteLine("WormBench - Synthetic workload for Transactinal Memory Systems");
            Console.WriteLine("Barcelona Supercpmputing Center - www.bscmsrc.eu");
            Console.WriteLine();
        }

        private static void PrintFooter()
        {
            Console.WriteLine();            
            Console.WriteLine("Done");
        }        

        private static void PrintUsage()
        {
            Console.WriteLine();
            Console.WriteLine("USAGE:");
            Console.WriteLine("WormBench.exe -t <int> -worms <string> -benchworld <string> -operations <string> [-demo] [-verify] [-log <string>] [-iterations <int>] [-timeout <int>]");
            Console.WriteLine();
            Console.WriteLine("OPTIONS:");
            Console.WriteLine("  - t: the number of the threads to create.");
            Console.WriteLine("  - worms: the worms initialization file.");            
            Console.WriteLine("  - benchworld: the name of the file that contains the benchworld.");
            Console.WriteLine("  - operations: the file containing the operations to be executed.");
            Console.WriteLine("  - demo: run visual demonstration.");
            Console.WriteLine("  - verify: verify that the execution was correct.");
            Console.WriteLine("  - log: a file to output logs; default the standard output.");
            Console.WriteLine("  - iterations: how many times to repeat the execution of the operations stream.");
            Console.WriteLine("  - timeout: specifies the timeout for the application; if the application exceed the timeout it exits properly.");
            Console.WriteLine();
            Console.WriteLine("EXAMPLE:");
            Console.WriteLine("1. > WormBench.exe -t 4 -worms WormsDir -benchworld BenchWorld.txt -operations MixedOpeations.txt");
            Console.WriteLine("     Execute WormBench by creating 4 threads and with worm ");
            Console.WriteLine("     configuration file located in Worms.txt, BenchWorld in ");
            Console.WriteLine("     BenchWorld.txt andthe operations contained in file");
            Console.WriteLine("     MixedOpeations.txt");
            Console.WriteLine();
            Console.WriteLine("2. > WormBench.exe -t 4 -worms WormsDir -benchworld BenchWorld.txt -operations MixedOpeations.txt -demo");
            Console.WriteLine("     Execute WormBench in Demo mode.");
            Console.WriteLine();
            Console.WriteLine("3. > WormBench.exe -t 4 -worms WormsDir -benchworld BenchWorld.txt -operations MixedOpeations.txt -verify");
            Console.WriteLine("     Execute WormBench and when finished verify that the.");
            Console.WriteLine("     execution was correct.");

        }

        private static void RunBenchmark()
        {
            Console.WriteLine("Benchmark started...");
            

            //
            // Spawn a thread for every worm and run one worm on the main thread.
            //
            for (int i = 0; i < Config.Worms.Length; i++ )
            {
                Config.Worms[i].Start();
            }

            parallelTimeStart = System.DateTime.Now.Ticks;

            //
            // Set the timeout. So that the threads exit normaly and not by enofrcing.
            //
            if (timeoutSeconds > 0)
            {
                TimeSpan waitTime = new TimeSpan(0, 0, 0, timeoutSeconds);               
                for (int i = 0; (i < Config.Worms.Length) && (waitTime > TimeSpan.Zero); i++)
                {
                    Config.Worms[i].Join(waitTime);

                    TimeSpan elapsedTime = new TimeSpan((int)(System.DateTime.Now.Ticks - parallelTimeStart));                    
                    waitTime = waitTime - elapsedTime;
                }

                //
                // Here we don't do synchornization before writing because we
                // assume that a concurent update and read will not affect the application semantics.
                //
                Config.IsTimedOut = true;
            }

            for (int i = 0; i < Config.Worms.Length; i++)
            {
                Config.Worms[i].Join();                
            }

            parallelTimeStop = System.DateTime.Now.Ticks;

            Console.WriteLine("Benchmark completed...");
        }
    }
}
